import os

from sqlalchemy import create_engine

from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.data.models.basemodel import BaseModel

DATABASE = os.environ.get('DATABASE', 'db.sqlite3')
DATABASEECHO = True if os.environ.get('DATABASEECHO', '0') =='1' else False


@singleton
class Engine:
    def __init__(self):
        self._engine = create_engine(f'sqlite:///{DATABASE}', echo=DATABASEECHO)
        BaseModel.metadata.create_all(self._engine)

    def get(self):
        return self._engine


from PySide6.QtWidgets import QGraphicsItemGroup

from mosamaticdesktop.data.file import File


class Layer:
    def __init__(self, name: str, index: int=-1, opacity: float=0.5, visible: bool=True) -> None:
        self._name = name
        self._index = index
        self._opacity = opacity
        self._visible = visible
        self._file = None

    def name(self) -> str:
        return self._name
    
    def index(self) -> int:
        return self._index
    
    def setIndex(self, index: int) -> None:
        self._index = index

    def opacity(self) -> float:
        return self._opacity
    
    def setOpacity(self, opacity) -> None:
        self._opacity = opacity

    def visible(self) -> bool:
        return self._visible
    
    def setVisible(self, visible) -> None:
        self._visible = visible

    def file(self) -> File:
        return self._file

    def setFile(self, file: File) -> None:
        self._file = file

    def createGraphicsItem(self) -> QGraphicsItemGroup:
        raise NotImplementedError('Not implemented')


import uuid

from PySide6.QtWidgets import QDockWidget

MENU_HEIGHT = 50


class DockWidget(QDockWidget):
    def __init__(self, title: str, parent=None) -> None:
        super(DockWidget, self).__init__(title, parent=parent)
        self.topLevelChanged.connect(self.toggleMaximize)
        self.setObjectName(str(uuid.uuid4()))

    def toggleMaximize(self):
        if self.isFloating():
            self.resize(self.parent().size())
            self.centerWindow()
        else:
            self.setFloating(False)

    def centerWindow(self):
        parentGeometry = self.parent().frameGeometry()
        geometry = self.frameGeometry()
        x = parentGeometry.x() + (parentGeometry.width() - geometry.width()) / 2
        y = parentGeometry.y() + (parentGeometry.height() - geometry.height()) / 2 + MENU_HEIGHT
        self.move(x, y)


import os

from PySide6.QtGui import QStandardItem, QFont

from mosamaticdesktop.data.fileset import FileSet


class FileSetItem(QStandardItem):
    def __init__(self, fileSet: FileSet) -> None:
        super(FileSetItem, self).__init__(fileSet.name())
        self._fileSet = fileSet
        deleted = False
        for file in self._fileSet.files():
            if not os.path.isfile(file.path()):
                deleted = True
                break
        if deleted:
            font = QFont()
            font.setItalic(True)
            self.setFont(font)
            self.setText(self._fileSet.name() + '[some or all files deleted]')

    def fileSet(self) -> FileSet:
        return self._fileSet


from typing import Any

from mosamaticdesktop.tasks.parameter import Parameter


class DescriptionParameter(Parameter):
    def __init__(self, name: str, description: str) -> None:
        super(DescriptionParameter, self).__init__(name=name, labelText=description, optional=True)

    def copy(self):
        return DescriptionParameter(
            name=self.name(), 
            description=self.labelText()
        )


import os
import zipfile

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.utils import createNameWithTimestamp
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class CreateArchiveTask(Task):
    def __init__(self) -> None:
        super(CreateArchiveTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Creates ZIP Archive From File Set'
        )
        self.addMultiFileSetParameter(
            name='inputFileSetNames',
            labelText='Input File Sets',
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def execute(self) -> None:

        # Get parameters needed for this task
        # inputFileSetName = self.parameter(name='inputFileSetName').value()
        inputFileSets = []
        inputFileSetNames = self.parameter('inputFileSetNames').value()
        for inputFileSetName in inputFileSetNames:
            inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
            inputFileSets.append(inputFileSet)
        inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)

        zipFileName = createNameWithTimestamp(outputFileSetName) + '.zip'
        outputZipFilePath = os.path.join(outputFileSetPath, zipFileName)

        step = 0
        files = []
        for inputFileSet in inputFileSets:
            files.extend(inputFileSet.files())
        nrSteps = len(files)
        with zipfile.ZipFile(outputZipFilePath, 'w') as zipObj:
            for file in files:
                zipObj.write(file.path(), arcname=os.path.basename(file.path()))
                self.updateProgress(step=step, nrSteps=nrSteps)
                step += 1

        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        LOGGER.info('Finished')



from typing import Any

from PySide6.QtWidgets import QWidget, QVBoxLayout

from mosamaticdesktop.tasks.parameter import Parameter

class ParameterWidget(QWidget):
    @classmethod
    def NAME(cls):
        return cls.__qualname__

    def __init__(self, parameter: Parameter, parent: QWidget=None) -> None:
        super(ParameterWidget, self).__init__(parent=parent)
        self._parameter = parameter
        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(5, 5, 5, 5)
        self._layout.setSpacing(5)
        self.setLayout(self._layout)

    def parameter(self) -> Parameter:
        return self._parameter

    def name(self) -> str:
        return self._parameter.name()

    def labelText(self) -> str:
        labelText = self._parameter.labelText()
        if not self.optional():
            labelText += '*'
        return labelText
    
    def optional(self) -> bool:
        return self._parameter.optional()
    
    def visible(self) -> bool:
        return self._parameter.visible()
    
    def defaultValue(self) -> Any:
        return self._parameter.defaultValue()
    
    def value(self) -> Any:
        return self._parameter.value()
    
    def setValue(self, value: Any) -> None:
        self._parameter.setValue(value)

    def layout(self) -> QVBoxLayout:
        return self._layout
    
    def copy(self):
        raise NotImplementedError()


from typing import Any, List

from mosamaticdesktop.tasks.parameter import Parameter


class OptionGroupParameter(Parameter):
    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, options: List[str]=[]) -> None:
        super(OptionGroupParameter, self).__init__(
            name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._options = options

    def options(self) -> List[str]:
        return self._options

    def copy(self):
        return OptionGroupParameter(
            name=self.name(), 
            labelText=self.labelText(), 
            optional=self.optional(), 
            visible=self.visible(), 
            defaultValue=self.defaultValue(),
            options=self.options(),
        )


import numpy as np

from PySide6.QtGui import QImage, QPixmap
from PySide6.QtWidgets import QGraphicsItemGroup, QGraphicsPixmapItem

from mosamaticdesktop.widgets.viewers.dicomviewer.layer import Layer
from mosamaticdesktop.utils import convertNumPyArrayToRgbQImage, AlbertaColorMap, readFromCache, writeToCache


class SegmentationLayer(Layer):
    def __init__(self) -> None:
        super(SegmentationLayer, self).__init__(name='numpy')
        self._qimage = None

    def data(self) -> np.array:
        if self.file():
            content = readFromCache(file=self.file())
            if not content:
                data = np.load(self.file().path())
                content = writeToCache(file=self.file(), fileObject=data)
            data = content.fileObject()
            return data
        return None

    def convertDataToQImage(self) -> QImage:
        return convertNumPyArrayToRgbQImage(numpyArray=self.data(), colorMap=AlbertaColorMap())

    def createGraphicsItem(self) -> QGraphicsItemGroup:
        group = QGraphicsItemGroup()
        if self.file() and not self._qimage:
            self._qimage = self.convertDataToQImage()
        if self._qimage:
            pixmap = QPixmap.fromImage(self._qimage)
            pixmapItem = QGraphicsPixmapItem(pixmap)
            pixmapItem.setOpacity(self.opacity())
            group.addToGroup(pixmapItem)
        return group


import os
import shutil
import numpy as np

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.utils import convertDicomToNumPyArray
from mosamaticdesktop.utils import convertNumPyArrayToPngImage
from mosamaticdesktop.utils import AlbertaColorMap
from mosamaticdesktop.utils import isNumPyFile
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class CreatePngFromSegmentationFileTask(Task):
    def __init__(self) -> None:
        super(CreatePngFromSegmentationFileTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description=f'Create PNGs From NumPy Files',
        )
        self.addFileSetParameter(
            name='inputFileSetName',
            labelText='Input File Set',
        )
        self.addPathParameter(
            name='outputFileSetPath',
            labelText='Output File Set Path',
        )
        self.addTextParameter(
            name='outputFileSetName',
            labelText='Output File Set Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputFileSet',
            labelText='Overwrite Output File Set',
            defaultValue=True,
        )

    def execute(self) -> None:

        # Get parameters needed for this task
        inputFileSetName = self.parameter(name='inputFileSetName').value()
        inputFileSet = self.dataManager().fileSetByName(name=inputFileSetName)
        outputFileSetPath = self.parameter('outputFileSetPath').value()
        outputFileSetName = self.parameter('outputFileSetName').value()
        if outputFileSetName is None:
            outputFileSetName = self.generateTimestampForFileSetName(name=inputFileSetName)
        overwriteOutputFileSet = self.parameter('overwriteOutputFileSet').value()
        outputFileSetPath = os.path.join(outputFileSetPath, outputFileSetName)
        if overwriteOutputFileSet:
            if os.path.isdir(outputFileSetPath):
                shutil.rmtree(outputFileSetPath)
        os.makedirs(outputFileSetPath, exist_ok=True)

        step = 0
        files = inputFileSet.files()
        nrSteps = len(files)
        for file in files:
            if isNumPyFile(file.path()):
                numpyArray = np.load(file.path())
                pngImageFileName = os.path.split(file.path())[1] + '.png'
                convertNumPyArrayToPngImage(
                    numpyArrayFilePathOrObject=numpyArray, 
                    colorMap=AlbertaColorMap(),
                    outputDirectoryPath=outputFileSetPath,
                    pngImageFileName=pngImageFileName,
                )
                LOGGER.info(f'{file.path()} created {pngImageFileName}')
            self.updateProgress(step=step, nrSteps=nrSteps)
            step += 1
        self.dataManager().createFileSet(fileSetPath=outputFileSetPath)
        LOGGER.info('Finished')


from typing import Any

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QVBoxLayout, QCheckBox

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget


class BooleanParameterWidget(ParameterWidget):
    def __init__(self, parameter: Parameter, parent: QWidget=None) -> None:
        super(BooleanParameterWidget, self).__init__(parameter=parameter, parent=parent)
        self._booleanCheckBox = None
        self.initUi()

    def initUi(self) -> None:
        self._booleanCheckBox = QCheckBox(self)
        self._booleanCheckBox.setText(self.parameter().labelText())
        if self.parameter().defaultValue() is not None:
            self._booleanCheckBox.setCheckState(Qt.Checked if self.parameter().defaultValue() else Qt.Unchecked)
            self.stateChanged(state=self._booleanCheckBox.checkState())
        self._booleanCheckBox.stateChanged.connect(self.stateChanged)
        self.layout().addWidget(self._booleanCheckBox)

    def stateChanged(self, state: int) -> None:
        self.parameter().setValue(True if state == Qt.Checked else False)


import os
import sys


class OperatingSystem:
    @staticmethod
    def isDarwin() -> bool:
        return sys.platform.startswith('darwin')
    
    @staticmethod
    def isWindows() -> bool:
        return sys.platform.startswith('win')
    
    @staticmethod
    def homeDirectory() -> str:
        if OperatingSystem.isDarwin():
            return os.environ['HOME']
        if OperatingSystem.isWindows():
            return os.environ['USERPROFILE']
        return None


from typing import Any

from PySide6.QtWidgets import QWidget, QLabel, QComboBox

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget
from mosamaticdesktop.data.datamanager import DataManager


class FileSetParameterWidget(ParameterWidget):
    class ComboBox(QComboBox):
        def __init__(self, parent: QWidget=None) -> None:
            super(FileSetParameterWidget.ComboBox, self).__init__(parent=parent)
            self._dataManager = DataManager()
            self.loadItems()

        def loadItems(self) -> None:
            self.addItem(None)
            fileSets = self._dataManager.fileSets()
            for fileSet in fileSets:
                self.addItem(fileSet.name())

        def showPopup(self) -> None:
            self.clear()
            self.loadItems()
            return super().showPopup()

    def __init__(self, parameter: Parameter, parent: QWidget=None) -> None:
        super(FileSetParameterWidget, self).__init__(parameter=parameter, parent=parent)
        if self.parameter().defaultValue() is not None:
            self.parameter().setValue(self.parameter().defaultValue())
        self.initUi()

    def initUi(self) -> None:
        label = QLabel(self.labelText())
        comboBox = FileSetParameterWidget.ComboBox(self)
        comboBox.currentTextChanged.connect(self.currentTextChanged)
        self.layout().addWidget(label)
        self.layout().addWidget(comboBox)

    def currentTextChanged(self, text: str) -> None:
        self.parameter().setValue(text)


import threading

from typing import Dict, List, Any

from PySide6.QtCore import QObject, Signal

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.descriptionparameter import DescriptionParameter
from mosamaticdesktop.tasks.labelparameter import LabelParameter
from mosamaticdesktop.tasks.filesetparameter import FileSetParameter
from mosamaticdesktop.tasks.multifilesetparameter import MultiFileSetParameter
from mosamaticdesktop.tasks.pathparameter import PathParameter
from mosamaticdesktop.tasks.filepathparameter import FilePathParameter
from mosamaticdesktop.tasks.textparameter import TextParameter
from mosamaticdesktop.tasks.integerparameter import IntegerParameter
from mosamaticdesktop.tasks.floatingpointparameter import FloatingPointParameter
from mosamaticdesktop.tasks.booleanparameter import BooleanParameter
from mosamaticdesktop.tasks.optiongroupparameter import OptionGroupParameter
from mosamaticdesktop.data.datamanager import DataManager
from mosamaticdesktop.logger import Logger
# from mosamaticdesktop.utils import createNameWithTimestamp

LOGGER = Logger()


class Task:
    class TaskSignal(QObject):
        progress = Signal(int)

    @classmethod
    def NAME(cls):
        # Returns class name of child classes
        return cls.__qualname__

    def __init__(self) -> None:
        self._thread = None
        self._parameters = {}
        self._dataManager = DataManager()
        self._signal = Task.TaskSignal()

    def dataManager(self) -> DataManager:
        return self._dataManager

    def signal(self) -> TaskSignal:
        return self._signal

    def start(self) -> None:
        self._thread = threading.Thread(target=self.run)
        self._thread.start()

    def run(self) -> None:
        self.execute()

    def execute(self) -> None:
        raise NotImplementedError()    

    def updateProgress(self, step: int, nrSteps: int) -> None:
        if self._signal:
            self._progress = int(((step + 1) / (nrSteps)) * 100)
            self._signal.progress.emit(self._progress)

    def cancel(self) -> None:
        self._thread.join()

    def parameters(self) -> List[Parameter]:
        return self._parameters.values()

    def parameter(self, name: str) -> Parameter:
        if name in self._parameters.keys():
            return self._parameters[name]
        return None
    
    def parameterValuesAsString(self) -> str:
        string = ''
        for name in self._parameters.keys():
            value = self._parameters[name].value()
            string += f'{name}="{value}", '
        return string[:-2]
    
    def updateParameters(self, parameters: List[Parameter]):
        for parameter in parameters:
            self._parameters[parameter.name()] = parameter
    
    def addDescriptionParameter(self, name: str, description: str) -> Parameter:
        parameter = DescriptionParameter(name=name, description=description)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addLabelParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = LabelParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addFileSetParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = FileSetParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter
    
    def addMultiFileSetParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = MultiFileSetParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addPathParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = PathParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addFilePathParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = FilePathParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addTextParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = TextParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addIntegerParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, minimum: int=0, maximum: int=100, step: int=1) -> Parameter:
        parameter = IntegerParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, minimum=minimum, maximum=maximum, step=step)
        LOGGER.info(f'TaskWidget.addIntegerParameter() value={parameter.name()}')
        self._parameters[parameter.name()] = parameter
        return parameter

    def addFloatingPointParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = FloatingPointParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter
    
    def addBooleanParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> Parameter:
        parameter = BooleanParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        self._parameters[parameter.name()] = parameter
        return parameter

    def addOptionGroupParameter(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None, options: List[str]=[]) -> Parameter:
        parameter = OptionGroupParameter(name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue, options=options)
        self._parameters[parameter.name()] = parameter
        return parameter



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QComboBox, QVBoxLayout, QPushButton

from mosamaticdesktop.widgets.dockwidget import DockWidget
from mosamaticdesktop.tasks.taskwidgetmanager import TaskWidgetManager
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class TaskDockWidget(DockWidget):
    def __init__(self, title: str) -> None:
        super(TaskDockWidget, self).__init__(title)
        self._tasksComboBox = None
        self._showTaskWidgetButton = None
        self._placeholderWidget = None
        self._taskWidgetManager = TaskWidgetManager()
        self._currentTaskName = None
        self.initUi()

    def initUi(self) -> None:
        self._tasksComboBox = QComboBox(self)
        self._tasksComboBox.addItem(None)
        for taskName in sorted(self._taskWidgetManager.taskNames()):
            self._tasksComboBox.addItem(taskName)
        self._tasksComboBox.currentIndexChanged.connect(self.currentIndexChanged)
        self._showTaskWidgetButton = QPushButton('Show Task Widget')
        self._showTaskWidgetButton.clicked.connect(self.showTaskWidget)
        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignTop)
        layout.addWidget(self._tasksComboBox)
        layout.addWidget(self._showTaskWidgetButton)
        widget = QWidget()
        widget.setLayout(layout)    
        self.setWidget(widget)
        self.setMinimumHeight(200)

    def showTaskWidget(self) -> None:
        if self._currentTaskName:
            taskWidget = self._taskWidgetManager.taskWidget(name=self._currentTaskName)
            taskWidget.show()

    def currentIndexChanged(self, index) -> None:
        taskName = self._tasksComboBox.itemText(index)
        if taskName:
            self._currentTaskName = taskName


import os

from typing import List, Dict

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.utils import ModuleLoader
from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


@singleton
class TaskWidgetManager:
    def __init__(self) -> None:
        self._taskWidgetTypes = self.loadTaskWidgetTypes()
        self._taskWidgets = {}

    def taskNames(self) -> List[str]:
        return list(self._taskWidgetTypes.keys())
    
    def taskWidgets(self) -> Dict[str, TaskWidget]:
        return self._taskWidgets
    
    def taskWidget(self, name) -> TaskWidget:
        if name in self._taskWidgetTypes.keys():
            if name not in self._taskWidgets.keys():
                self._taskWidgets[name] = self._taskWidgetTypes[name]()
            return self._taskWidgets[name]
        LOGGER.error(f'TaskWidgetManager: task widget for task {name} does not exist')
        return None
    
    def loadTaskWidgetTypes(self):
        moduleDirectoryPath = os.path.dirname(os.path.realpath(__file__))
        LOGGER.info(f'TaskWidgetManager: loading task widget types from {moduleDirectoryPath}...')
        taskWidgets = ModuleLoader.loadModuleClasses(
            moduleDirectoryPath=moduleDirectoryPath,
            moduleBaseClass=TaskWidget,
            fileNameEndsWith='taskwidget.py',
        )
        return taskWidgets


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.totalsegmentatortask.totalsegmentatortask import TotalSegmentatorTask


class TotalSegmentatorTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(TotalSegmentatorTaskWidget, self).__init__(taskType=TotalSegmentatorTask)
    
    def validate(self) -> None:
        pass


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.createpngfromsegmentationfiletask.createpngfromsegmentationfiletask import CreatePngFromSegmentationFileTask


class CreatePngFromSegmentationFileTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(CreatePngFromSegmentationFileTaskWidget, self).__init__(taskType=CreatePngFromSegmentationFileTask)
    
    def validate(self) -> None:
        pass


import os

from PySide6.QtGui import QStandardItem, QFont

from mosamaticdesktop.data.file import File


class FileItem(QStandardItem):
    def __init__(self, file: File) -> None:
        super(FileItem, self).__init__(file.name())
        self._file = file
        if not os.path.isfile(self._file.path()):
            font = QFont()
            font.setItalic(True)
            self.setFont(font)
            self.setText(self._file.name() + '[deleted]')

    def file(self) -> File:
        return self._file


from typing import Any

from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel, QFileDialog, QLineEdit, QVBoxLayout, QPushButton

from mosamaticdesktop.tasks.parameter import Parameter
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget
from mosamaticdesktop.utils import Configuration


class FilePathParameterWidget(ParameterWidget):
    def __init__(self, parameter: Parameter, parent: QWidget=None) -> None:
        super(FilePathParameterWidget, self).__init__(parameter=parameter, parent=parent)
        if self.parameter().defaultValue() is not None:
            self.parameter().setValue(self.parameter().defaultValue())
        self._pathLineEdit = None
        self._settings = Configuration().qSettings()
        self.initUi()

    def initUi(self) -> None:
        button = QPushButton('Select File Path...', self)
        button.setFixedWidth(150)
        button.clicked.connect(self.showFileDialog)
        self._pathLineEdit = QLineEdit(self)
        self.layout().addWidget(QLabel(self.parameter().labelText()))
        self.layout().addWidget(self._pathLineEdit)
        self.layout().addWidget(button)

    def showFileDialog(self) -> None:
        lastFilePath = self._settings.value('lastFileOpened')
        filePath, _ = QFileDialog.getOpenFileName(self, 'Select File Path', lastFilePath)
        if filePath:
            self._pathLineEdit.setText(filePath)
            self.parameter().setValue(value=filePath)
            self._settings.setValue('lastFileOpened', filePath)


from mosamaticdesktop.singleton import singleton
from mosamaticdesktop.data.filecontent import FileContent


@singleton
class FileContentCache:
    def __init__(self) -> None:
        self._cache = {}

    def get(self, id: str) -> FileContent:
        if self.has(id):
            return self._cache[id]
        return None
    
    def has(self, id: str) -> bool:
        return id in self._cache.keys()
        
    def add(self, fileContent: FileContent) -> None:
        self._cache[fileContent.id()] = fileContent

    def remove(self, id: str) -> None:
        if id in self._cache.keys():
            del self._cache[id]

    def clear(self) -> None:
        self._cache = {}


import uuid

from typing import List
from sqlalchemy import Column, String
from sqlalchemy.orm import Mapped, relationship

from mosamaticdesktop.data.models.basemodel import BaseModel
from mosamaticdesktop.utils import createNameWithTimestamp


class FileSetModel(BaseModel):
    __tablename__ = '_filesetmodel'
    id: Mapped[int] = Column('_id', String, primary_key=True, default=lambda: str(uuid.uuid4()), unique=True, nullable=False)
    name: Mapped[str] = Column('_name', String(256), nullable=False)
    path: Mapped[str] = Column('_path', String(1024), nullable=False)
    fileModels: Mapped[List['FileModel']] = relationship('FileModel', back_populates='fileSetModel', cascade='all, delete-orphan')



from typing import List

from mosamaticdesktop.data.models.filesetmodel import FileSetModel
from mosamaticdesktop.data.file import File


class FileSet:
    def __init__(self, model: FileSetModel) -> None:
        self._id = model.id
        self._name = model.name
        self._path = model.path
        
        self._files = []
        for fileModel in model.fileModels:
            file = File(model=fileModel)
            self._files.append(file)

    def id(self) -> str:
        return self._id

    def name(self) -> str:
        return self._name
    
    def setName(self, name: str) -> None:
        self._name = name
    
    def path(self) -> str:
        return self._path
    
    def addFile(self, file: File) -> None:
        self._files.append(file)

    def files(self) -> List[File]:
        return self._files

    def nrFiles(self) -> int:
        return len(self._files)


import os
import shutil

from PySide6.QtCore import Qt
from PySide6.QtGui import QWheelEvent
from PySide6.QtWidgets import QWidget, QGraphicsView, QGraphicsScene, QSlider, QCheckBox, QGridLayout, QLabel, QPushButton, QHBoxLayout, QMessageBox
from PySide6.QtWidgets import QFileDialog

from mosamaticdesktop.widgets.viewers.dicomviewer.dicomlayer import DicomLayer
from mosamaticdesktop.widgets.viewers.dicomviewer.dicominfolayer import DicomInfoLayer
from mosamaticdesktop.widgets.viewers.dicomviewer.segmentationlayer import SegmentationLayer
from mosamaticdesktop.widgets.viewers.dicomviewer.taglayer import TagLayer
from mosamaticdesktop.widgets.viewers.dicomviewer.filesetcombobox import FileSetComboBox
from mosamaticdesktop.utils import isDicomFile
from mosamaticdesktop.data.fileset import FileSet
from mosamaticdesktop.data.datamanager import DataManager
from mosamaticdesktop.data.file import File
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class DicomViewer(QWidget):
    class MyGraphicsView(QGraphicsView):
        def __init__(self, parent=None):
            super(DicomViewer.MyGraphicsView, self).__init__(parent)

        def wheelEvent(self, event: QWheelEvent) -> None:
            event.ignore()

    def __init__(self) -> None:
        super(DicomViewer, self).__init__()
        self._graphicsWidget = None
        self._graphicsView = None
        self._scene = None
        self._layerTuples = []
        self._windowLevel = 50
        self._windowWidth = 400
        # UI
        self._settingsWidget = None
        self._layerTupleSlider = None
        self._segmentationFileOpacityLabel = None
        self._segmentationFileOpacitySlider = None
        self._tagFileOpacityLabel = None
        self._tagFileOpacitySlider = None
        self._dicomFileInfoOpacityLabel = None
        self._dicomFileInfoOpacitySlider = None
        self._inputFileSetComboBox = None
        # Miscellaneous
        self._fileSetName = None
        self._fullScan = False
        self._fullScanCheckBox = None
        self._currentLayerTupleIndex = 0
        self._exportDirectory = None
        self._dataManager = DataManager()
        self._layout = None
        self.initUi()

    def initUi(self) -> None:
        self.initGraphicsView()
        self.initSliders()
        self.initFileSetComboBox()
        self._fullScanCheckBox = QCheckBox('Full Scan')
        self._fullScanCheckBox.setCheckState(Qt.Checked if self._fullScan else Qt.Unchecked)
        self._fullScanCheckBox.stateChanged.connect(self.fullScanCheckBoxStateChanged)
        updateViewerButton = QPushButton('Update Viewer')
        updateViewerButton.clicked.connect(self.inputFileSetChanged)
        selectExportDirectoryButton = QPushButton('Select Export Directory')
        selectExportDirectoryButton.clicked.connect(self.selectExportDirectory)
        exportCurrentImageButton = QPushButton('Export Current Image to File')
        exportCurrentImageButton.clicked.connect(self.exportCurrentImage)
        layout = QGridLayout()
        layout.addWidget(self._graphicsView, 0, 0, 1, 2)
        layout.addWidget(QLabel('Image'), 1, 0)
        layout.addWidget(self._layerTupleSlider, 1, 1)
        layout.addWidget(QLabel('Opacity Segmentation'), 2, 0)
        layout.addWidget(self._segmentationFileOpacitySlider, 2, 1)
        layout.addWidget(QLabel('Opacity TAG'), 3, 0)
        layout.addWidget(self._tagFileOpacitySlider, 3, 1)
        layout.addWidget(QLabel('Input File Set'), 4, 0)
        layout.addWidget(self._inputFileSetComboBox, 4, 1)
        layout.addWidget(self._fullScanCheckBox, 5, 1)
        layout.addWidget(updateViewerButton, 6, 1)
        buttonLayout = QHBoxLayout()
        buttonLayout.addWidget(selectExportDirectoryButton)
        buttonLayout.addWidget(exportCurrentImageButton)
        layout.addLayout(buttonLayout, 7, 1)
        self.setLayout(layout)

    def initGraphicsView(self) -> None:
        self._graphicsView = DicomViewer.MyGraphicsView(self)
        self._scene = QGraphicsScene(self)
        item = self._scene.addText('DICOM Viewer')
        item.setDefaultTextColor(Qt.blue)
        self._graphicsView.setScene(self._scene)
        self._graphicsView.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._graphicsView.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

    def initSliders(self) -> None:
        self._layerTupleSlider = QSlider(Qt.Horizontal, self)
        self._layerTupleSlider.setRange(0, 100)
        self._layerTupleSlider.valueChanged.connect(self.currentLayerTupleIndexChanged)
        self._segmentationFileOpacitySlider = QSlider(Qt.Horizontal, self)
        self._segmentationFileOpacitySlider.setRange(0, 100)
        self._segmentationFileOpacitySlider.setValue(50)
        self._segmentationFileOpacitySlider.valueChanged.connect(self.segmentationFileOpacityChanged)
        self._segmentationFileOpacitySlider.setVisible(False)
        self._tagFileOpacitySlider = QSlider(Qt.Horizontal, self)
        self._tagFileOpacitySlider.setRange(0, 100)
        self._tagFileOpacitySlider.setValue(50)
        self._tagFileOpacitySlider.valueChanged.connect(self.tagFileOpacityChanged)
        self._tagFileOpacitySlider.setVisible(False)

    def initFileSetComboBox(self) -> None:
        self._inputFileSetComboBox = FileSetComboBox(self)

    def findSegmentationFileForDicomFile(self, dicomFile: File, fileSet: FileSet) -> File:
        for file in fileSet.files():
            if file.name() == dicomFile.name() + '.seg.npy':
                return file
        return None

    def findTagFileForDicomFile(self, dicomFile: File, fileSet: FileSet) -> File:
        for file in fileSet.files():
            if file.name() == dicomFile.name() + '.tag':
                return file
            elif file.name() == dicomFile.name()[:-4] + '.tag':
                return file
            else:
                pass
        return None
    
    # def startSetInputFileSet(self, fileSet: FileSet):
    #     worker = threading.Thread(target=self.setInputFileSet, args=(fileSet, ))
    #     worker.start()

    def setInputFileSet(self, fileSet: FileSet) -> None:
        self._fileSetName = fileSet.name()
        self._segmentationFileOpacitySlider.setVisible(False)
        self._tagFileOpacitySlider.setVisible(False)
        step = 0
        nrSteps = len(fileSet.files())
        for file in fileSet.files():
            layerTuple = [None, None, None, None] # dicom, numpy, tag, info
            if isDicomFile(filePath=file.path()):
                dicomLayer = DicomLayer()
                dicomLayer.setFile(file=file)
                dicomLayer.setWindowLevelAndWidth(windowLevel=self._windowLevel, windowWidth=self._windowWidth)
                layerTuple[0] = dicomLayer
                segmentationFile = self.findSegmentationFileForDicomFile(dicomFile=file, fileSet=fileSet)
                if segmentationFile:
                    segmentationLayer = SegmentationLayer()
                    segmentationLayer.setFile(file=segmentationFile)
                    layerTuple[1] = segmentationLayer
                    self._segmentationFileOpacitySlider.setVisible(True)
                tagFile = self.findTagFileForDicomFile(dicomFile=file, fileSet=fileSet)
                if tagFile:
                    tagLayer = TagLayer()
                    tagLayer.setFile(file=tagFile)
                    tagLayer.setShape([dicomLayer.data().Rows, dicomLayer.data().Columns])
                    layerTuple[2] = tagLayer
                    self._tagFileOpacitySlider.setVisible(True)
                dicomInfoLayer = DicomInfoLayer()
                dicomInfoLayer.setFile(file=file)
                dicomInfoLayer.setInstanceNumber(instanceNumber=dicomLayer.data().InstanceNumber)
                layerTuple[3] = dicomInfoLayer
                self._layerTuples.append(layerTuple)
            step += 1
        if self._fullScan:
            LOGGER.info('Sorting full scan images by instance number...')
            self._layerTuples = sorted(self._layerTuples, key=lambda layerTuple: layerTuple[0].data().InstanceNumber)
        self._layerTupleSlider.setRange(0, len(self._layerTuples) - 1)
        self.displayLayerTuple(self._currentLayerTupleIndex)

    def currentLayerTupleIndexChanged(self, index) -> None:
        if index > 0 and index < len(self._layerTuples):
            self._currentLayerTupleIndex = index
            self.displayLayerTuple(self._currentLayerTupleIndex)

    def segmentationFileOpacityChanged(self, value) -> None:
        if self._layerTuples:
            for layerTuple in self._layerTuples:
                layerTuple[1].setOpacity(value / 100.0)
            self.displayLayerTuple(self._currentLayerTupleIndex)

    def tagFileOpacityChanged(self, value) -> None:
        if self._layerTuples:
            for layerTuple in self._layerTuples:
                layerTuple[2].setOpacity(value / 100.0)
            self.displayLayerTuple(self._currentLayerTupleIndex)

    def fullScanCheckBoxStateChanged(self, state) -> None:
        self._fullScan = True if state == Qt.Checked else False

    def inputFileSetChanged(self) -> None:
        text = self._inputFileSetComboBox.currentText()
        if text:
            self.clearData()
            inputFileSet = self._dataManager.fileSetByName(text)
            # self.startSetInputFileSet(fileSet=inputFileSet)
            self.setInputFileSet(fileSet=inputFileSet)

    def selectExportDirectory(self) -> None:
        self._exportDirectory = QFileDialog.getExistingDirectory(self, 'Select Export Directory')
        LOGGER.info(f'Selected export directory: {self._exportDirectory}')

    def exportCurrentImage(self) -> None:
        if self._currentLayerTupleIndex >= 0 and len(self._layerTuples) > 0:
            if self._exportDirectory:
                file = self._layerTuples[self._currentLayerTupleIndex][0].file()
                if file:
                    fileName = os.path.split(file.path())[1]
                    exportedFilePath = os.path.join(self._exportDirectory, f'{self._fileSetName}-{fileName}')
                    shutil.copy(file.path(), exportedFilePath)
                    LOGGER.info(f'Exported file {file.path()} to export directory')
                    QMessageBox.information(self, 'Success', f'Succesfully saved file {file.name()} to {self._exportDirectory}')
                    

    def wheelEvent(self, event) -> None:
        delta = event.angleDelta().y()
        if delta > 0 and self._currentLayerTupleIndex > 0:
            self._currentLayerTupleIndex -= 1
        elif delta < 0 and self._currentLayerTupleIndex < len(self._layerTuples) - 1:
            self._currentLayerTupleIndex += 1
        # Temporarily block slider events
        self._layerTupleSlider.valueChanged.disconnect(self.currentLayerTupleIndexChanged)
        self._layerTupleSlider.setValue(self._currentLayerTupleIndex)
        self._layerTupleSlider.valueChanged.connect(self.currentLayerTupleIndexChanged)
        # Display image
        self.displayLayerTuple(self._currentLayerTupleIndex)

    def displayLayerTuple(self, index: int) -> None:
        if len(self._layerTuples) > 0 and index < len(self._layerTuples):
            layerTuple = self._layerTuples[index]
            self._scene.clear()
            self._scene.addItem(layerTuple[0].createGraphicsItem())
            if layerTuple[1]:
                self._scene.addItem(layerTuple[1].createGraphicsItem())
            if layerTuple[2]:
                self._scene.addItem(layerTuple[2].createGraphicsItem())
            self._scene.addItem(layerTuple[3].createGraphicsItem())

    def clearData(self) -> None:
        self._scene.clear()
        self._layerTuples = []
        item = self._scene.addText('DICOM Viewer')
        item.setDefaultTextColor(Qt.blue)



from PySide6.QtCore import Qt
from PySide6.QtWidgets import QGraphicsTextItem, QGraphicsItemGroup

from mosamaticdesktop.widgets.viewers.dicomviewer.layer import Layer


class DicomInfoLayer(Layer):
    def __init__(self) -> None:
        super(DicomInfoLayer, self).__init__(name='info')
        self._instanceNumber = -1

    def setInstanceNumber(self, instanceNumber: int) -> None:
        self._instanceNumber = instanceNumber

    def createGraphicsItem(self) -> QGraphicsItemGroup:
        y = 10
        group = QGraphicsItemGroup()
        if self.file():
            filePathItem = QGraphicsTextItem(self.file().path())
            filePathItem.setDefaultTextColor(Qt.white)
            filePathItem.setPos(10, y)
            y += 20
            group.addToGroup(filePathItem)
        if self._instanceNumber > -1:
            instanceNumberItem = QGraphicsTextItem(f'Instance Number: {self._instanceNumber}')
            instanceNumberItem.setDefaultTextColor(Qt.white)
            instanceNumberItem.setPos(10, y)
            y += 20
            group.addToGroup(instanceNumberItem)
        return group


def singleton(cls):
    _instances = {}

    def instance(*args, **kwargs) -> cls:
        if cls not in _instances:
            _instances[cls] = cls(*args, **kwargs)
        return _instances[cls]
    
    return instance


from pydicom.datadict import keyword_dict

from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.checkdicomheadertask.checkdicomheadertask import CheckDicomHeaderTask


class CheckDicomHeaderTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(CheckDicomHeaderTaskWidget, self).__init__(taskType=CheckDicomHeaderTask)

    def validate(self) -> None:
        requiredAttributes = self.taskParameterWidget('requiredAttributes').parameter().value()
        if requiredAttributes != '':
            attributes = [x.strip() for x in requiredAttributes.split(',')]
            for attribute in attributes:
                if attribute not in keyword_dict:
                    self.showValidationError('requiredAttributes', f'Attribute {attribute} is not a valid DICOM attribute')
                    return


import os

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout

from mosamaticdesktop.widgets.dockwidget import DockWidget
from mosamaticdesktop.widgets.viewers.dicomviewer.dicomviewer import DicomViewer
from mosamaticdesktop.data.datamanager import DataManager


class MainViewerDockWidget(DockWidget):
    def __init__(self, title: str) -> None:
        super(MainViewerDockWidget, self).__init__(title)
        self._viewer = None
        self._dataManager = DataManager()
        self.initUi()

    def initUi(self) -> None:
        self._viewer = DicomViewer()
        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignTop)
        layout.addWidget(self._viewer)
        widget = QWidget()
        widget.setLayout(layout)    
        self.setWidget(widget)


from typing import Any, List, Dict

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QPushButton, QProgressBar, QProgressBar, QDialog, QMessageBox

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.tasks.parameterwidget import ParameterWidget
from mosamaticdesktop.tasks.descriptionparameter import DescriptionParameter
from mosamaticdesktop.tasks.descriptionparameterwidget import DescriptionParameterWidget
from mosamaticdesktop.tasks.labelparameter import LabelParameter
from mosamaticdesktop.tasks.labelparameterwidget import LabelParameterWidget
from mosamaticdesktop.tasks.filesetparameter import FileSetParameter
from mosamaticdesktop.tasks.filesetparameterwidget import FileSetParameterWidget
from mosamaticdesktop.tasks.multifilesetparameter import MultiFileSetParameter
from mosamaticdesktop.tasks.multifilesetparameterwidget import MultiFileSetParameterWidget
from mosamaticdesktop.tasks.pathparameter import PathParameter
from mosamaticdesktop.tasks.pathparameterwidget import PathParameterWidget
from mosamaticdesktop.tasks.filepathparameter import FilePathParameter
from mosamaticdesktop.tasks.filepathparameterwidget import FilePathParameterWidget
from mosamaticdesktop.tasks.textparameter import TextParameter
from mosamaticdesktop.tasks.textparameterwidget import TextParameterWidget
from mosamaticdesktop.tasks.integerparameter import IntegerParameter
from mosamaticdesktop.tasks.integerparameterwidget import IntegerParameterWidget
from mosamaticdesktop.tasks.floatingpointparameter import FloatingPointParameter
from mosamaticdesktop.tasks.floatingpointparameterwidget import FloatingPointParameterWidget
from mosamaticdesktop.tasks.booleanparameter import BooleanParameter
from mosamaticdesktop.tasks.booleanparameterwidget import BooleanParameterWidget
from mosamaticdesktop.tasks.optiongroupparameter import OptionGroupParameter
from mosamaticdesktop.tasks.optiongroupparameterwidget import OptionGroupParameterWidget
from mosamaticdesktop.logger import Logger

LOGGER = Logger()


class TaskWidget(QDialog):
    @classmethod
    def NAME(cls):
        # Returns class name of child classes (and strips of last 6 characters to get the task name itself)
        return cls.__qualname__[:-6]

    def __init__(self, taskType: Task) -> None:
        super(TaskWidget, self).__init__()
        self._taskType = taskType
        self._progressBar = QProgressBar(self)
        self._parameterWidgets = self.createTaskParameterWidgetsFromTask(taskType=self._taskType)
        self._task = None
        self._startButton = None
        self._cancelButton = None
        self._closeButton = None
        self.initUi()

    def initUi(self) -> None:
        layout = QVBoxLayout()
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setSpacing(5)
        layout.setAlignment(Qt.AlignTop)
        for parameterName in self._parameterWidgets.keys():
            layout.addWidget(self._parameterWidgets[parameterName])
        layout.addWidget(self._progressBar)
        layout.addWidget(self.createButtonsWidget())
        self.setLayout(layout)
        self.setFixedWidth(500)
        self.setWindowTitle(self.__class__.__name__)

    def createTaskParameterWidgetsFromTask(self, taskType: Task) -> Dict[str, ParameterWidget]:
        widgets = {}
        task = taskType()
        for parameter in task.parameters():
            if isinstance(parameter, DescriptionParameter):
                widgets[parameter.name()] = DescriptionParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, LabelParameter):
                widgets[parameter.name()] = LabelParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, FileSetParameter):
                widgets[parameter.name()] = FileSetParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, MultiFileSetParameter):
                widgets[parameter.name()] = MultiFileSetParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, PathParameter):
                widgets[parameter.name()] = PathParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, FilePathParameter):
                widgets[parameter.name()] = FilePathParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, TextParameter):
                widgets[parameter.name()] = TextParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, IntegerParameter):
                widgets[parameter.name()] = IntegerParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, FloatingPointParameter):
                widgets[parameter.name()] = FloatingPointParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, BooleanParameter):
                widgets[parameter.name()] = BooleanParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, IntegerParameter):
                widgets[parameter.name()] = IntegerParameterWidget(parameter=parameter, parent=self)
            elif isinstance(parameter, OptionGroupParameter):
                widgets[parameter.name()] = OptionGroupParameterWidget(parameter=parameter, parent=self)
        return widgets

    def createButtonsWidget(self) -> None:
        self._startButton = QPushButton('Start')
        self._startButton.setEnabled(True)
        self._startButton.setFocus()
        self._startButton.clicked.connect(self.start)
        self._cancelButton = QPushButton('Cancel')
        self._cancelButton.setEnabled(False)
        self._cancelButton.clicked.connect(self.cancel)
        self._closeButton = QPushButton('Close')
        self._closeButton.setEnabled(True)
        self._closeButton.clicked.connect(self.close)
        buttonsLayout = QHBoxLayout()
        buttonsLayout.addWidget(self._startButton)
        buttonsLayout.addWidget(self._cancelButton)
        buttonsLayout.addWidget(self._closeButton)
        buttonsLayout.setAlignment(Qt.AlignRight)
        buttonsWidget = QWidget()
        buttonsWidget.setLayout(buttonsLayout)
        return buttonsWidget
    
    def validate(self) -> None:
        raise NotImplementedError()
    
    def showValidationError(self, parameterName: str, message: str) -> None:
        QMessageBox.critical(self, f'Validation Error for parameter "{parameterName}"', message)

    def start(self) -> None:
        for parameter in self._parameterWidgets.values():
            if not parameter.optional():
                if parameter.value() is None or parameter.value() == '':
                    QMessageBox.critical(self, 'Error', f'Parameter {parameter.name()} cannot be empty!')
                    return
        self._task = self._taskType()
        self._task.updateParameters(self._parameterWidgets.values())
        self._task.signal().progress.connect(self.progress)
        self._task.start()
        self._startButton.setEnabled(False)
        self._cancelButton.setEnabled(True)
        self._closeButton.setEnabled(False)
        self._progressBar.setValue(0)

    def progress(self, progress: int) -> None:
        self._progressBar.setValue(progress)
        if progress >= 100:
            self._cancelButton.setEnabled(False)
            self._startButton.setEnabled(True)
            self._closeButton.setEnabled(True)            

    def cancel(self) -> None:
        if self._task:
            self._task.cancel()
            self._cancelButton.setEnabled(False)
            self._startButton.setEnabled(True)
            self._closeButton.setEnabled(True)
            self._progressBar.setValue(0)

    def show(self):
        return self.exec_()

    def close(self) -> None:
        self.accept()



import os
import shutil
import nibabel as nib
import warnings

from totalsegmentator.python_api import totalsegmentator

from mosamaticdesktop.tasks.task import Task
from mosamaticdesktop.logger import Logger
from mosamaticdesktop.utils import createNameWithTimestamp
from mosamaticdesktop.tasks.totalsegmentatortask.checksegmentation import CheckSegmentation
from mosamaticdesktop.tasks.totalsegmentatortask.sliceselector import SliceSelector
from mosamaticdesktop.utils import currentTimeInSeconds, elapsedSeconds

warnings.filterwarnings('ignore', 'Invalid value for VR UI')

LOGGER = Logger()

ROIS = [
    'vertebrae_S1', 'vertebrae_C1', 'vertebrae_C2', 'vertebrae_C3', 'vertebrae_C4', 'vertebrae_C5', 'vertebrae_C6', 'vertebrae_C7', 
    'vertebrae_L1', 'vertebrae_L2', 'vertebrae_L3', 'vertebrae_L4', 'vertebrae_L5', 'vertebrae_T1', 'vertebrae_T2', 'vertebrae_T3', 
    'vertebrae_T4', 'vertebrae_T5', 'vertebrae_T6', 'vertebrae_T7', 'vertebrae_T8', 'vertebrae_T9', 'vertebrae_T10', 'vertebrae_T11', 
    'vertebrae_T12'
]


class TotalSegmentatorTask(Task):
    """ This task requires Torch + CUDA. Install using pip:
    pip3 install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu118
    """
    def __init__(self) -> None:
        super(TotalSegmentatorTask, self).__init__()
        self.addDescriptionParameter(
            name='description',
            description='Extracts anatomical ROIs from list of full CT scans'
        )
        self.addPathParameter(
            name='rootDirectoryPath',
            labelText='Root Directory of CT Scans (each scan in separate subdirectory)'
        )
        self.addPathParameter(
            name='outputDirectoryPath',
            labelText='Output Directory for Segmentations'
        )
        self.addOptionGroupParameter(
            name='vertebra',
            labelText='Select Vertebral Level',
            options=ROIS,
        )
        self.addOptionGroupParameter(
            name='device',
            labelText='Select Device (CPU or GPU)',
            options=['GPU', 'CPU'],
            defaultValue='GPU',
        )
        self.addBooleanParameter(
            name='fast',
            labelText='Enable Fast Calculation (for CPU only)',
            defaultValue=False,
        )
        self.addBooleanParameter(
            name='qualityCheck',
            labelText='Perform Quality Check After Segmentation',
            defaultValue=True,
        )
        self.addTextParameter(
            name='outputDirectoryName',
            labelText='Output Directory Name',
            optional=True,
        )
        self.addBooleanParameter(
            name='overwriteOutputDirectory',
            labelText='Overwrite Output Directory',
            defaultValue=True,
        )

    def execute(self) -> None:

        # Get root directory path containing sub-directories for each CT scan
        rootDirectoryPath = self.parameter('rootDirectoryPath').value()
        if rootDirectoryPath:
            outputDirectoryName = self.parameter('outputDirectoryName').value()
            if outputDirectoryName is None:
                outputDirectoryName = createNameWithTimestamp(name=outputDirectoryName)
            outputDirectoryPath = self.parameter('outputDirectoryPath').value()
            outputDirectoryPath = os.path.join(outputDirectoryPath, outputDirectoryName)
            LOGGER.info(f'Output directory path: {outputDirectoryPath}')

            # Get vertebra
            vertebra = self.parameter('vertebra').value()
            LOGGER.info(f'vertebra = {vertebra}')

            # Quality check yes/no
            qualityCheck = self.parameter('qualityCheck').value()
            LOGGER.info(f'Quality check = {qualityCheck}')

            # Get device and fast options. Disable fast if we're using the GPU
            device = self.parameter('device').value().lower()
            fast = self.parameter('fast').value()
            if device == 'gpu' and fast:
                fast = False
            LOGGER.info(f'device = {device}, fast = {fast}')

            # Check whether to overwrite previous output directory
            overwriteOutputDirectory = self.parameter('overwriteOutputDirectory').value()
            LOGGER.info(f'Overwrite output fileset: {overwriteOutputDirectory}')
            if overwriteOutputDirectory:
                if os.path.isdir(outputDirectoryPath):
                    shutil.rmtree(outputDirectoryPath)
            os.makedirs(outputDirectoryPath, exist_ok=True)

            selectedSlicesDirectoryPath = os.path.join(outputDirectoryPath, 'slices')
            if os.path.isdir(selectedSlicesDirectoryPath):
                shutil.rmtree(selectedSlicesDirectoryPath)
            os.makedirs(selectedSlicesDirectoryPath, exist_ok=False)

            # Each CT scan's files should be in a separate subdirectory inside the root directory
            step = 0
            nrSteps = 0
            for scanDirectoryName in os.listdir(rootDirectoryPath):
                scanDirectoryPath = os.path.join(rootDirectoryPath, scanDirectoryName)
                if os.path.isdir(scanDirectoryPath):
                    nrSteps += 1

            startTimeTotal = currentTimeInSeconds()
            for scanDirectoryName in os.listdir(rootDirectoryPath):
                scanDirectoryPath = os.path.join(rootDirectoryPath, scanDirectoryName)
                if os.path.isdir(scanDirectoryPath):
                    startTime = currentTimeInSeconds()
                    LOGGER.info(f'Running TotalSegmentator on scan directory {scanDirectoryPath}...')
                    outputScanDirectoryPath = os.path.join(outputDirectoryPath, scanDirectoryName)
                    os.makedirs(outputScanDirectoryPath, exist_ok=True)

                    # Run Total Segmentator twice, once to get individual NIFTI files for each ROI. 
                    # And once to store all labels in a single NIFTI. This volume will be used for quality checking
                    totalsegmentator(
                        scanDirectoryPath, outputScanDirectoryPath, fast=fast, device=device)
                    roiFilePath = os.path.join(outputScanDirectoryPath, vertebra + '.nii.gz')

                    ok = True
                    if qualityCheck:
                        totalsegmentator(
                            scanDirectoryPath, outputScanDirectoryPath, fast=fast, device=device, ml=True)
                        segmentationFilePath = os.path.join(outputDirectoryPath, scanDirectoryName + '.nii') # No .gz extension!
                        segmentation = nib.load(segmentationFilePath)
                        checker = CheckSegmentation(segmentation=segmentation, scanName=scanDirectoryName)
                        ok = checker.execute()

                    if ok:
                        # Get requested ROI and select DICOM slice running through it
                        roi = nib.load(roiFilePath)
                        selector = SliceSelector(roi=roi, volume=segmentation, dicomDirectory=scanDirectoryPath)
                        output_files = selector.execute()
                        if len(output_files) > 0:
                            LOGGER.info(f'Found median slice {output_files[0]} for {vertebra}')
                            selectedSlice = os.path.join(selectedSlicesDirectoryPath, scanDirectoryName + '-' + vertebra + '-' + os.path.split(output_files[0])[1])
                            shutil.copyfile(output_files[0], selectedSlice)
                            LOGGER.info(f'Elapsed time after one scan: {elapsedSeconds(startTime)} seconds')
                        else:
                            LOGGER.error(f'Output of slice selector contains more than one file')
                    else:
                        LOGGER.error(f'Error checking segmentation')

                    self.updateProgress(step=step, nrSteps=nrSteps)
                    step += 1

            self.dataManager().createFileSet(fileSetPath=selectedSlicesDirectoryPath)
            LOGGER.info(f'Total elapsed time: {elapsedSeconds(startTimeTotal)} seconds')

            LOGGER.info('Finished')
            LOGGER.info(f'Results can be found in: {outputDirectoryPath}')


from typing import Any

from mosamaticdesktop.tasks.parameter import Parameter


class FloatingPointParameter(Parameter):
    def __init__(self, name: str, labelText: str, optional: bool=False, visible: bool=True, defaultValue: Any=None) -> None:
        super(FloatingPointParameter, self).__init__(
            name=name, labelText=labelText, optional=optional, visible=visible, defaultValue=defaultValue)
        if self.defaultValue() is not None:
            self.setValue(self.defaultValue())
    
    def copy(self):
        return FloatingPointParameter(
            name=self.name(), 
            labelText=self.labelText(), 
            optional=self.optional(), 
            visible=self.visible(), 
            defaultValue=self.defaultValue()
        )


from mosamaticdesktop.tasks.taskwidget import TaskWidget
from mosamaticdesktop.tasks.musclefatsegmentationtask.musclefatsegmentationtask import MuscleFatSegmentationTask


class MuscleFatSegmentationTaskWidget(TaskWidget):
    def __init__(self) -> None:
        super(MuscleFatSegmentationTaskWidget, self).__init__(taskType=MuscleFatSegmentationTask)
    
    def validate(self) -> None:
        pass


from PySide6.QtWidgets import QProgressBar

from mosamaticdesktop.singleton import singleton


@singleton
class ProgressBarWidget(QProgressBar):
    def __init__(self) -> None:
        super(ProgressBarWidget, self).__init__()